import multiprocessing


import pickle
from generator import Generator, Parameters, SurrogateDataset
from reconstruction import Reconstruction
#from surrogate import Surrogate
from one_dim_flow_surrogate import Surrogate
from optimizer import Optimizer
from matplotlib import pyplot as plt
import time
import torch
import numpy as np
import os

from tools import plot_evolution_with_panels

def reset_weights(m):
    '''
    Try resetting model weights.
    '''
    for layer in m.children():
        if hasattr(layer, 'reset_parameters'):
            print(f'Reset trainable parameters of layer = {layer}')
            layer.reset_parameters()




if __name__ == "__main__":
    #check i context has already been set
    if not hasattr(multiprocessing, 'get_start_method'):
        multiprocessing.set_start_method('spawn')

    outpath = 'TESTING_FLOW_ECAL_HCAL'
    
    os.system('mkdir  '+outpath)
    outpath += '/'
    os.system('cp *.py '+outpath)

    test = False

    def write_to_log(*args):
        #make a string with white spaces between the arguments out of the arguments
        message = ' '.join([str(a) for a in args])
        print(message)
        with open(outpath+'/log.txt', 'a') as f:
            f.write(message+'\n')
    #remove old log file
    os.system('rm -f '+outpath+'/log.txt')


    divide = 10 if test else 1

    #so far these are only layer thicknesses
    start_pars = {f'thickness_absorber_{i}': .1 for i in range(100)}
    start_pars.update({f'thickness_scintillator_{i}': 0.1 for i in range(100)})

    box = np.array(len(start_pars)*[1.]) #will be adjusted to some extent by optimizer
    #start parameters from run 8
    #start_pars = {'thickness_absorber_0': 0.7642903, 'thickness_absorber_1': 10.469371, 'thickness_scintillator_0': 30.585306, 'thickness_scintillator_1': 22.256506}
    
    
    gen = Generator(box, 
                    Parameters(parameters = start_pars),
                    n_vars = 24, # 30, DEBUG
                    n_events_per_var = 100//divide, #400
                    particles=[['gamma',0.22], ['pi+',2.11]],
                    max_cores = 24)#,['pi+',2.11]])#0)
    
    write_to_log('will generate', gen.n_vars * gen.n_events_per_var * len(gen.particles) , 'events each step')
    
    # reconstruction
    reco_model = Reconstruction(gen.n_parameters, 
                           gen.n_parameters_per_sensor * gen.n_sensors, 
                           gen.n_target_parameters)
    
    surrogate_model = Surrogate(gen.n_parameters, 
                           gen.n_target_parameters + gen.n_context_parameters, #context
                           gen.n_target_parameters,  #reco
                           n_time_steps=100,
                           #fix_xmin = -1e-3
                           ) # FLOW_ADJUSTMENT: this is the minimum value of the loss, the small negative value is to avoid numerical problems
    
    optimizer = Optimizer(gen, surrogate_model, reco_model, gen.parameters,
                          constraints= {'length': 200}, #125 adjusted to also fit some hadron part
                          surrogate_predicts_loss=True) #FLOW_ADJUSTMENT: this uses the flow model that predicts the loss
    
    n_epochs_pre = 50//divide
    n_epochs_main = 50//divide
    
    parameters = gen.parameters

    #if saved evolution exists, load it
    import os.path
    if False and os.path.isfile('evolution.npy'):
        evolution = np.load('evolution.npy').tolist()
        print('EVOLUTION:',evolution)
        parameters = evolution[-1]
    
    evolution = []
    def save_evolution(parameters, losses, reco_losses, reco_val_loss, reco_model, means, std):
        '''
        Add to evolution list and save it with pickle.
        Save it in human readable format as well.
        '''
        # Translate and update parameters
        ps_dict = gen.translate_parameters(parameters)
        ps = {k: ps_dict[k] for k in ps_dict.keys()}
        
        # Append all relevant data to evolution
        step_data = {
            'parameters': ps,
            'losses': losses,
            'reco_losses': reco_losses,
            'reco_val_loss': reco_val_loss,
            'reco_model': reco_model.state_dict(),
            'gen': {'means': means, 'stds': std}
        }
        evolution.append(step_data)
        
        # Save evolution to pickle
        with open(outpath + 'evolution.pkl', 'wb') as f:
            pickle.dump(evolution, f)

        plot_evolution_with_panels(outpath + 'evolution.pkl', outpath)



    #save_evolution(parameters, 1., 1.)
    #exit()
    
    def make_check_plot(surr_out, reco_out, true_in, evolution, outname="" , take_face_value = False):
        '''
        make response plots for all three and save them, use evolution size as step index for output file name
        '''
        #if any are torch tensors detach and numpy them
        if hasattr(surr_out, 'detach'):
            surr_out = surr_out.detach().cpu().numpy()
        if hasattr(reco_out, 'detach'):
            reco_out = reco_out.detach().cpu().numpy()
        if hasattr(true_in, 'detach'):
            true_in = true_in.detach().cpu().numpy()


        if take_face_value:
            reco_resp = reco_out
            surr_resp = surr_out
        else:
            reco_resp = (reco_out - true_in) / true_in
            surr_resp = (surr_out - true_in) / true_in
    
        #range = np.min([np.min(reco_resp), np.min(surr_resp)]), np.max([np.max(reco_resp), np.max(surr_resp)])

        if len(outname) == 0:
            outname = str(len(evolution))
        
        range = (-1,1)
        if take_face_value:
            range = (0, max(np.max(reco_resp), np.max(surr_resp)))
        #use fixed bins from the range
        bins = np.linspace(range[0], range[1], 31)

        
        #put all nan anf inf values below minimum range by 10% of total range
        soor = np.logical_or(np.isnan(surr_resp), np.isinf(surr_resp))
        reco_resp = np.where(soor, range[0] - 0.1*(range[1]-range[0]), reco_resp)

        plt.figure(figsize=(10,10))
        plt.hist(reco_resp.flatten(), bins=bins, range=range, alpha = 0.5, label='G4+reco')
        plt.hist(surr_resp.flatten(), bins=bins, range=range, alpha = 0.5, label='E2E surrogate')
        plt.text(0.15, 0.8, 'step '+outname, transform=plt.gca().transAxes)
        

        plt.xlabel('response')
        if take_face_value:
            plt.xlabel('loss')
        plt.legend()
        plt.savefig(outpath+'response_'+outname+'_lin.png')
        try:
            plt.yscale('log')
            plt.savefig(outpath+'response_'+outname+'_log.png')
        except:
            pass
        plt.close()

        #the same plot but using log of the responses
        plt.figure(figsize=(10,10))
        #re-determine the range based on log(reco_resp) etc
        range = (np.min([np.min(np.log(reco_resp+1e-19)), np.min(np.log(surr_resp))]), 
                 np.max([np.max(np.log(reco_resp+1e-19)), np.max(np.log(surr_resp))]))
        bins = np.linspace(range[0], range[1], 31)

        plt.hist(np.log(reco_resp.flatten()), bins=bins, alpha = 0.5, label='G4+reco')
        plt.hist(np.log(surr_resp.flatten()), bins=bins, alpha = 0.5, label='E2E surrogate')
        plt.text(0.15, 0.8, 'step '+outname, transform=plt.gca().transAxes)
        plt.savefig(outpath+'log_response_'+outname+'_lin.png')
        plt.close()

    #pre-train reco and surrogate model
    def pre_train():
        ds = gen.create_torch_dataset(parameters, random_scaler=1.2)
        reco_model.to('cuda')
        write_to_log('pre-training 0')
        reco_model.train_model(ds, batch_size=128, n_epochs= 2, lr=0.03)
        write_to_log('pre-training 1')
        reco_model.train_model(ds, batch_size=256, n_epochs= n_epochs_pre, lr=0.01)
        write_to_log('pre-training 2')
        reco_model.train_model(ds, batch_size=512, n_epochs= n_epochs_pre, lr=0.001)
        write_to_log('pre-training 3')
        reco_model.train_model(ds, batch_size=2048, n_epochs= n_epochs_pre, lr=0.001)
        #this is a normalised quantity - at some point only expose the un-normalised ones, but also needs adjustments for surrogate dataset
        reco_result,reco_loss = reco_model.apply_model_in_batches(ds, batch_size=128)

        reco_model.to('cpu')
    
        # FLOW_ADJUSTMENT (or more general): now we use the reco loss loss as surrogate target!
        surrogate_dataset = SurrogateDataset(ds, reco_loss.detach().cpu().numpy())#important to detach here
        if hasattr(surrogate_model, 'is_flow_model') and surrogate_model.is_flow_model:
            # FLOW_ADJUSTMENT: the flow needs an x_min and x_max range to follow the dataset
            # this sets the initial range, which is fine for loss prediction
            surrogate_model.adjust_xmin_xmax(reco_loss.detach().cpu().numpy())
            write_to_log("Adjusted x_min and x_max to follow the reco_loss, xmin,xmax:", 
                         surrogate_model.x_min,surrogate_model.x_max, 'in log space this is',
                         surrogate_model.model.xmin, surrogate_model.model.xmax)

        surrogate_model.to('cuda')
        surrogate_model.model.unfreeze_batchnorm()
        write_to_log('surr pre-training 0')
        surrogate_model.train_model(surrogate_dataset, batch_size=256, n_epochs= n_epochs_pre, lr=0.001)
        write_to_log('surr pre-training 1')
        surrogate_model.train_model(surrogate_dataset, batch_size=512, n_epochs= n_epochs_pre, lr=0.0005)
        write_to_log('surr pre-training 2')
        surrogate_model.model.freeze_batchnorm()
        surrogate_model.train_model(surrogate_dataset, batch_size=2048, n_epochs= n_epochs_pre, lr=0.0001)
        write_to_log('surr pre-training 3')
        surrogate_model.train_model(surrogate_dataset, batch_size=4096, n_epochs= 2*n_epochs_pre, lr=1e-5)
        #these are un-normalised quantities
        # FLOW_ADJUSTMENT
        surr_reco_loss_out, reco_out, true_in = surrogate_model.apply_model_in_batches(surrogate_dataset, batch_size=2048)
        # FLOW_ADJUSTMENT: 
        make_check_plot(surr_reco_loss_out, reco_loss, true_in, evolution, outname = "pretrain", take_face_value= True)
    pre_train()
    #exit() #DEBUG

    best_surrogate_loss = 1e10
    def update_best_surrogate_loss(loss):
        # FLOW_ADJUSTMENT: the flow loss can have negative values so
        # this function does not work. check if the surrogate is the flow model
        if hasattr(surrogate_model, 'is_flow_model') and surrogate_model.is_flow_model:
            return True
        
        global best_surrogate_loss
        if loss < best_surrogate_loss:
            best_surrogate_loss = loss
            return True
        return loss < 10. * best_surrogate_loss

    covariance_pcas = []
    
    for i in range(1000): #just an example

        write_to_log("Step", i)
        ds = gen.create_torch_dataset(parameters) # full variations plus a tiny bit
        write_to_log("Reconstruction training")
        surrogate_model.to('cpu')
        reco_model.to('cuda')
        reco_model.train_model(ds, batch_size=1024, n_epochs= n_epochs_main, lr=0.0001)
        reco_model.train_model(ds, batch_size=1024, n_epochs= n_epochs_main, lr=1e-5)
        reco_result, reco_loss = reco_model.apply_model_in_batches(ds, batch_size=128)
        reco_model.to('cpu') #free up gpu memory
        reco_loss = reco_loss.detach().cpu().numpy()
        reco_result = reco_result.detach().cpu().numpy()
    
        #create the surrogate dataset; FLOW_ADJUSTMENT: use reco_loss as target!
        surrogate_dataset = SurrogateDataset(ds, reco_loss)#important to detach here
        write_to_log(" |-> Eval reconstruction loss", np.mean(reco_loss), "+-", np.std(reco_loss))
        
        write_to_log("Surrogate training")
        if hasattr(surrogate_model, 'is_flow_model') and surrogate_model.is_flow_model:
            # FLOW_ADJUSTMENT: the flow needs an x_min and x_max range to follow the dataset
            # these are computed as twice the range using moving averages (so transfer learning works)
            # xmin is anyway set to zero, but adjust xmax here
            surrogate_model.adjust_xmin_xmax(reco_loss)
            write_to_log("Adjusted x_min and x_max to follow the reco_loss, xmin, xmax:", surrogate_model.x_min, surrogate_model.x_max,
                         'in log space this is', (surrogate_model.model.xmin), (surrogate_model.model.xmax))
            
        surrogate_model.to('cuda')
        surrogate_model.model.unfreeze_batchnorm()
        surrogate_model.train_model(surrogate_dataset, batch_size=256, n_epochs= n_epochs_main, lr=0.003) #get the surrogate model to a good state
        surrogate_model.model.freeze_batchnorm()
        surrogate_model.train_model(surrogate_dataset, batch_size=512, n_epochs= n_epochs_main, lr=0.0001) #get the surrogate model to a good state
        sl = surrogate_model.train_model(surrogate_dataset, batch_size=2048, n_epochs= n_epochs_main, lr=1e-5) #get the surrogate model to a good state
    
        while not update_best_surrogate_loss(sl):
            write_to_log("Surrogate retraining")
            pre_train()
            surrogate_model.train_model(surrogate_dataset, batch_size=256, n_epochs= n_epochs_main//5, lr=0.002)
            surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.002)
            surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.0003)
            sl = surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.0001)

        ## check if things worked
        surr_reco_loss_out, reco_out, true_in = surrogate_model.apply_model_in_batches(surrogate_dataset, batch_size=512)
        make_check_plot(surr_reco_loss_out, reco_loss, true_in, evolution, take_face_value= True)
    
        write_to_log("Optimization")
        #apply the surrogate model
        updated_detector_parameters, optimal, o_loss = optimizer.optimize(surrogate_dataset, 
                                                                          batch_size=512, n_epochs=500, lr=0.003, 
                                                                          add_constraints=True)#some noise
        write_to_log("Updated detector parameters: ", gen.translate_parameters(updated_detector_parameters), "; not interrupted", optimal, "; change", updated_detector_parameters - parameters)
        write_to_log("Optimizer loss", o_loss)
        write_to_log("Optimizer maximum bad gradient retries", optimizer.retries)


        #use the difference of old and updated parameters to define a direction in parameter space
        #and increase the box size in that direction
        #gen.box_size += updated_detector_parameters - parameters
        # make sure the size does not become too small in any dimension
        #gen.box_size = np.maximum(gen.box_size, 0.1 * np.ones_like(gen.box_size))

        if np.isnan(o_loss) or np.isnan(updated_detector_parameters).any() or np.isinf(o_loss) or np.isinf(updated_detector_parameters).any():
            write_to_log("NaN loss, restarting in step",i)
            #add a line to a log in the output directory that this happened
            surrogate_model.apply(reset_weights)
            reco_model.apply(reset_weights)
            #reset optimizer.optimizer status
            optimizer.reset_optimizer()
            pre_train()
            sl = 1e6 #reset surrogate loss too to avoid infinite loop
            continue


        #just do that every 5 steps
        reco_val_loss = None
        if i % 5 == 0:
            #create a small test dataset on the optimised parameters
            write_to_log('Creating test dataset, step', i)
            test_events = 1000
            if test:
                test_events = 100
            #use the old parameters here for which the model was optimised mostly
            ds = gen.create_torch_dataset(parameters, random_scaler=1e-6, n_events_per_var= test_events//gen.n_vars)#total of 1k events only
            #apply it to the reco model and record the val loss
            reco_result, reco_val_loss = reco_model.apply_model_in_batches(ds, batch_size=128)
            reco_val_loss = reco_val_loss.detach().mean().cpu().numpy()
        #
        # step more in the direction. 
        parameters = updated_detector_parameters + 0.5*(updated_detector_parameters - parameters)


        # for bookkeeping, append the generator box_covariance to a text file
        with open(outpath+'box_covariance.txt', 'a') as f:
            f.write(str(gen.box_covariance) + '\n')
        
        # also perform a quick PCA of the box covariance and plot the first two principal components
        # but for all entries of covariance_pcas
        pca = np.linalg.eig(gen.box_covariance)
        covariance_pcas.append(pca)
        plt.plot(range(len(covariance_pcas)), [p[0][0] for p in covariance_pcas])
        plt.plot(range(len(covariance_pcas)), [p[0][1] for p in covariance_pcas])
        plt.xlabel('Step')
        plt.ylabel('1st & 2nd principal component')
        plt.savefig(outpath+f'pca.png')#just overwrite
        plt.close()


        
        save_evolution(updated_detector_parameters, o_loss, np.mean(reco_loss), reco_val_loss, reco_model, gen.means, gen.stds) #save the optimised parameters here, not the ones for the next box generation
        #exit()

    
    

    
