

import numpy as np
import pandas as pd
import torch
from torch.utils.data import Dataset, DataLoader
import time
import multiprocessing
import os

#check if G4Calo is the right package, in other words if run_batch can be imported
# if not, the following will fail
try:
    from G4Calo import run_batch
except ImportError:
    print("G4Calo package not found. Make sure to use the jkiesele/minicalosim:latest docker image. The ml_base image does not work anymore.")


# make geant quiet

'''
A small helper class to map indices to names and vice versa that can be accessed as a dict or array.
just some boiler plate that will help in the future
'''
class IndexMapper(object):
    def __init__(self, names : list):
        self.names = names
        self.indices = {}
        for i in range(len(names)):
            self.indices[names[i]] = i
    
    def get_index(self, name : str):
        #check if exists
        if name not in self.indices.keys():
            raise ValueError("Name {} not in indices!".format(name))
        return self.indices[name]
    
    def get_name(self, index : int):
        #check if exists
        if index not in range(len(self.names)):
            raise ValueError("Index {} not in names!".format(index))
        return self.names[index]
    
    def map_to_dict(self, array : np.ndarray):
        return {self.names[i] : array[i] for i in range(len(array))}
    
    def map_to_array(self, dictionary : dict):
        return np.array([dictionary[name] for name in self.names])


class Parameters(object):
    '''
    Class to store the parameters of the geometry.
    Can return and take parameters as numpy array but is configured with using explicit parameter names.
    Only certain parameters are allowed to ensure that the connection between array and parameter name is consistent.
    For now, the allowed parameters are:
    - layer_thickness, alternating between absorber and scintillator

    If materials etc are added, the mapping from float to material name has to be added here.
    '''
    def __init__(self, N :int = 0, parameters : dict = {}):
        
        assert N == 0 or len(parameters) == 0, "Either specify N or parameters, not both!"
        assert N != 0 or len(parameters) != 0, "Either specify N or parameters!"

        #rewrite this to use the IndexMapper class
        names = []
        if len(parameters) == 0:
            names = ['thickness_absorber_'+str(i) for i in range(N//2)]
            names += ['thickness_scintillator_'+str(i) for i in range(N//2)]
            vals = np.random.uniform(0.1, 1, size=N)
        else:
            names = list(parameters.keys())
            vals = np.array(list(parameters.values())) #guaranteede to be same order?
            N = len(names)

        self.mapper = IndexMapper(names)
        self.parameters = vals
        self.N = N

    def translate_parameters(self, parameters : np.ndarray):
        '''
        Translates the parameters from the given array to the internal array.
        '''
        assert len(parameters) == self.N, "Wrong number of parameters!"
        return self.mapper.map_to_dict(parameters)
    
    def set_parameters(self, parameters : np.ndarray):
        '''
        Sets the parameters from a numpy array.
        '''
        assert len(parameters) == self.N, "Wrong number of parameters!"
        self.parameters = parameters

    def get_descriptor(self, parameters : np.ndarray = None, desc_class = None):
        '''
        Returns a GeometryDescriptor from the given parameters.
        Strictly takes a dict as input to ensure that the parameter names are consistent.
        '''
        if parameters is None:
            parameters = self.parameters

        # get dict representation
        parameters = self.mapper.map_to_dict(parameters) #takes a few ms, but worth it for clarity
        cw = desc_class()
        for i in range(self.N//2):
            absth = max([1e-3, parameters['thickness_absorber_'+str(i)]])
            cw.addLayer(absth, "G4_Pb", False)
            senth = max([1e-3, parameters['thickness_scintillator_'+str(i)]])
            #cw.addLayer(senth, "G4_POLYSTYRENE",True,1)
            if i==0:
                cw.addLayer(senth, "G4_PbWO4",True,1)
            else:
                cw.addLayer(senth, "G4_POLYSTYRENE",True,1) #DEBUG
            #cw.addLayer(senth, "G4_POLYSTYRENE",True,1) #DEBUG
        return cw

        
class CaloDataset(Dataset):

    def __init__(self, sensor_array, detector_array, target_array, context_array, means=None, stds=None):
        assert len(sensor_array) == len(detector_array) == len(target_array), "Arrays have different lengths!"
        #all numpy arrays
        assert isinstance(sensor_array, np.ndarray) and isinstance(detector_array, np.ndarray) and isinstance(target_array, np.ndarray), "Arrays are not numpy arrays!"
        self.sensor_array = sensor_array
        self.detector_array = detector_array
        self.target_array = target_array
        self.context_array = context_array
        # apply means and std if given
        if means is not None:
            self.sensor_array = (self.sensor_array - means[0]) / stds[0]
            # keep detector unnormalised
            self.detector_array = (self.detector_array - means[1]) / stds[1]

            #a normalised target is important for the surrogate to work given the scheduling we have here
            self.target_array = (self.target_array - means[2]) / stds[2]
            self.context_array = (self.context_array - means[3]) / stds[3]

            self.c_means = [torch.tensor(a).to('cuda') for a in means]
            self.c_stds = [torch.tensor(a).to('cuda') for a in stds] 

        self.filter_infs_and_nans()

    def filter_infs_and_nans(self):
        '''
        Removes all events that contain infs or nans.
        '''
        mask = np.ones(len(self.sensor_array), dtype=bool)
        for i in range(len(self.sensor_array)):
            if np.any(np.isinf(self.sensor_array[i])) or np.any(np.isnan(self.sensor_array[i])):
                mask[i] = False
        self.sensor_array = self.sensor_array[mask]
        self.detector_array = self.detector_array[mask]
        self.target_array = self.target_array[mask]
        self.context_array = self.context_array[mask]

    def unnormalise_target(self, target):
        '''
        receives back the physically meaningful target from the normalised target
        '''
        return target * self.c_stds[2] + self.c_means[2]
    
    def normalise_target(self, target):
        '''
        normalises the target
        '''
        return (target - self.c_means[2]) / self.c_stds[2]
    
    def unnormalise_detector(self, detector):
        '''
        receives back the physically meaningful detector from the normalised detector
        '''
        return detector * self.c_stds[1] + self.c_means[1]
    
    def normalise_detector(self, detector):
        '''
        normalises the detector
        '''
        return (detector - self.c_means[1]) / self.c_stds[1]
        
    def __len__(self):
        return len(self.sensor_array)
    
    def __getitem__(self, idx):
        return self.detector_array[idx] ,self.sensor_array[idx], self.target_array[idx]
    

class SurrogateDataset(CaloDataset):
    '''
    This dataset is initialised with an existing CaloDataset and copies most of its behaviour and data.
    However, it adds an additional array that contains the output of the reconstruction model for each event.

    Returns in iterator:
    - detector parameters
    - true inputs
    - reco output
    '''
    def __init__(self, calo_dataset, reco_output):
        super(SurrogateDataset, self).__init__(calo_dataset.sensor_array, 
                                               calo_dataset.detector_array, 
                                               calo_dataset.target_array,
                                               calo_dataset.context_array)
        self.reco_output = reco_output
        #move means and std over
        self.c_means = calo_dataset.c_means
        self.c_stds = calo_dataset.c_stds
        assert len(self.reco_output) == len(self.sensor_array), "Reco output and sensor array have different lengths!"

    def __getitem__(self, idx):
        return self.detector_array[idx], self.target_array[idx], self.context_array[idx], self.reco_output[idx]
            
#just a thin wrapper 
class Generator(object):
    def __init__(self, 
                 box_size: np.ndarray, 
                 par_desc: Parameters,  #nominal parameters
                 n_vars = 10,
                 n_events_per_var = 1000,
                 particles=[['gamma', 0.22],['pi+', 0.211]],
                 energy_range = [1,20],
                 max_cores = 24
                 ):
        '''
        The box size describes the range of the parameter variations that are performed in 
        one iteration of random_generate. The box size is an array of the same length as the
        parameter list. The box size is added to the nominal parameter value to get the upper
        limit of the parameter range. The lower limit is the nominal parameter value minus the
        box size.

        workflow:
        while not converged:
            set parmeters
            generate
            train reco and surrogate
            while is_local:
                (update parameters)
                is_local?

        '''
        assert box_size.shape == par_desc.parameters.shape, "Box size and parameters have different shapes!"
        self.box_size = box_size
        # make the box actually a covariance matrix
        self.box_covariance = np.diag(box_size**2)

        self.nominal_parameters = None
        self.par_desc = par_desc

        self.n_vars = n_vars
        self.n_events_per_var = n_events_per_var
        self.energy_range = energy_range

        self.sensor_parameters = ['sensor_energy', 'sensor_x', 'sensor_y', 'sensor_z', 'sensor_dx', 'sensor_dy', 'sensor_dz', 'sensor_layer']
        self.target_parameters = ['true_energy']
        self.context_parameters = ['true_pid']
        self.detector_parameters = ['detector_parameters']
        self.particles = particles
        self.max_cores = max_cores

        # will be roughly determined from the first 
        self.means = None
        self.stds = None
    
    def reset(self):
        self.nominal_parameters = None
        self.means = None
        self.stds = None

    @property
    def n_parameters(self):
        return len(self.par_desc.parameters)
    
    @property
    def n_sensors(self):
        return self.n_parameters // 2
    
    @property
    def n_parameters_per_sensor(self):
        return len(self.sensor_parameters)
    
    @property
    def n_target_parameters(self):
        return len(self.target_parameters)
    
    @property
    def n_context_parameters(self):
        return len(self.context_parameters)

    @property
    def parameters(self):
        return self.par_desc.parameters
    
    def translate_parameters(self, parameters : np.ndarray):
        '''
        Translates the parameters from the given array to the internal array.
        '''
        assert len(parameters) == self.n_parameters, "Wrong number of parameters!"
        return self.par_desc.translate_parameters(parameters)

    def is_local(self, parameters : np.ndarray, scaler = 1.):
        '''
        Returns true if the parameters are within the box size of the nominal parameters.
        '''
        if self.nominal_parameters is None:
            return False
        # check if the parmeters are within one sigma of the multivariate gaussian defined by box_covariance
        # do not use box_size here but really the covariance
        diff = parameters - self.nominal_parameters
        return np.dot(diff, np.dot(np.linalg.inv(self.box_covariance), diff)) < scaler

    
    
    def random_parameters(self, parameters : np.ndarray, scaler = 1.):
        '''
        Generates a random set of parameters around the nominal parameters.
        use a gaussian distribution for now, defined by box_covariance
        '''
        return np.random.multivariate_normal(parameters, self.box_covariance * scaler)
            
        
        return np.random.normal(parameters, scaler * np.array(self.box_size))
        
    
    def random_energy(self, scaler = 1.):
        '''
        Generates a random energy within the energy range, uniform distribution
        '''
        return np.random.uniform(self.energy_range[0], self.energy_range[1])

    def generate(self, parameters : np.ndarray, scaler = 1.):
        '''
        generates a dataframe with different variations of the parameters within the box size
        sets central parameters to parameters
        '''
        assert len(parameters) == len(self.par_desc.parameters), "Wrong number of parameters!"
        self.nominal_parameters = parameters

        from tools import ThreadingPool
        n_cores_pool = min(self.n_vars, self.max_cores)
        # create a pool of threads
        pool = ThreadingPool(n_cores_pool)
        # add the threads
        for i in range(self.n_vars):
            pool.add_task(self.generate_single, self.random_parameters(parameters, scaler))

        # run the threads
        pool.run()
        # get the results
        results = pool.get_results()
        # concatenate the results
        return pd.concat(results)
        
    def generate_single(self, parameters=None):#, result_queue=None):
        '''
        generates a dataframe with one set of parameters.
        this should probably be moved to the Parameters class, given that there are explicit mapping dependencies
        '''
        # import only in fork
        print('starting new generate single')

        from  G4Calo import GeometryDescriptor, run_batch

        print('running single')
        
        if parameters is None:
            parameters = self.parameters
        
        dfs = []
        for p in self.particles:
            #since we are going for no_mp we need to create a new descriptor for each process.
            print('running particle', p)
            cw = self.par_desc.get_descriptor(parameters, GeometryDescriptor)
            df = run_batch(cw, self.n_events_per_var, p[0], 1., 100, ) #no_mp=True)
            df = df.assign(true_pid=np.array( len(df) * [p[1]], dtype='float32'))
            dfs.append(df)
        df = pd.concat(dfs)
        tiled_pars = len(df)*[np.array(parameters) ] # np.tile(pars[np.newaxis, ...], [len(df), 1])
        # add the array as columns to the dataframe
        df = df.assign(detector_parameters=tiled_pars)
        
        print('done with generating single')
        return df
        result_queue.put(df)
        
        return 0

    


    def create_torch_dataset(self, parameters=None, random_scaler = 1., means=None, stds = None, n_events_per_var=None):
        # create a torch data loader from the dataframe
        # the dataframe has the following columns:
        # - true_energy [a scalar per event]
        # - sensor_energy [a vector per event]
        # - sensor_x,y,z [each a vector per event]
        # - sensor_dx,dy,dz [each a vector per event]
        # - sensor_layer [each a vector per event]
    
        # the data loader should return a tuple of input to the model and truth target
        # the input is one vector of length n_detector_parameters + n_input_parameters, and should always contain the detector parameters
        # as the first entries
        # the values for n_detector_parameters and n_input_parameters can be inferred from the dataframe
        # n_input_parameters is simply the number of sensors times the parameters per sensor (here 8)
    
        # infer the lengths
        if parameters is None:
            parameters = self.parameters #original parmeters
        
        if n_events_per_var is not None:
            #save old
            old_n_events_per_var = self.n_events_per_var
            self.n_events_per_var = n_events_per_var
        df = self.generate(parameters, random_scaler)
        if n_events_per_var is not None:
            #reset
            self.n_events_per_var = old_n_events_per_var
        print('done with generating')
        
        # create simple array data out of dataframe (will sit in memory, can be improved at some point)
        # this is the input to the data loader
    
    
        sensor_array = np.array([df[par].to_numpy() for par in self.sensor_parameters], dtype='float32') #.transpose()
        #transpose axis 0 and 1 
        sensor_array = np.swapaxes(sensor_array, 0, 1)
        #flatt to nevents x -1
        sensor_array = np.reshape(sensor_array, (len(sensor_array), -1))

        target_array = np.array([df[par].to_numpy() for par in self.target_parameters], dtype='float32').transpose()

        #this is kinda ugly. would be better to find a smarter way to add it to the dataframe in the first place
        detector_array = np.array([df[par].to_numpy() for par in self.detector_parameters])[0]
        detector_array = np.concatenate(detector_array,axis=0)
        detector_array = np.array(detector_array, dtype='float32')
        detector_array = np.reshape(detector_array, (len(target_array), -1))

        context_array = np.array([df[par].to_numpy() for par in self.context_parameters], dtype='float32').transpose()
        
        if means is None:
            self.means = [np.mean(sensor_array, axis=0), np.mean(detector_array, axis=0), np.mean(target_array, axis=0), np.mean(context_array, axis=0)]
        else:
            self.means = means
        if stds is None:
            self.stds = [np.std(sensor_array, axis=0)+1e-3, np.std(detector_array, axis=0)+1e-3, np.std(target_array, axis=0)+1e-3, np.std(context_array, axis=0)+1e-3]
        else:
            self.stds = stds
        # data ready, now create the torch Dataset
        
        ds = CaloDataset(sensor_array, detector_array, target_array, context_array, self.means, self.stds)
        return ds


def test():
    #test stuff
    gen = Generator(np.array([0.1, 0.1, 0.1, 0.1]), Parameters(4),
                    n_vars = 20,
                    n_events_per_var = 40)
    ds = gen.create_torch_dataset(np.array([0.5, 0.5, 0.5, 0.5]))
    
    dl = DataLoader(ds,  batch_size=3, shuffle=True)

    for i, (detector_parameters,sensor_parameters, target_parameters) in enumerate(dl):
        print(i, detector_parameters, sensor_parameters, target_parameters)

#test()
