#!/usr/bin/env python
# coding=utf-8

#########################################################################
# This program is free software: you can redistribute it and/or modify  #
# it under the terms of the version 3 of the GNU General Public License #
# as published by the Free Software Foundation.                         #
#                                                                       #
# This program is distributed in the hope that it will be useful, but   #
# WITHOUT ANY WARRANTY; without even the implied warranty of            #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      #
# General Public License for more details.                              #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program. If not, see <http://www.gnu.org/licenses/>.  #
#                                                                       #
# Written by and Copyright (C) Francois Fleuret                         #
# Contact <francois.fleuret@idiap.ch> for comments & bug reports        #
#########################################################################

import math
from math import pi
import random

import numpy as np
import torch
import torchvision
from torch import nn, autograd
from torch.nn import functional as F
import torch
from torch.utils.data import DataLoader
import matplotlib.pyplot as plt
import matplotlib.collections as mc

def set_seed(seed):
    if seed >= 0:
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

#set_seed(0)

######################################################################

def phi(x):
    p, std = 0.3, 0.2
    mu = (1 - p) * torch.exp(LogProba((x - 0.5) / std, math.log(1 / std))) + \
              p  * torch.exp(LogProba((x + 0.5) / std, math.log(1 / std)))
    return mu

def sample_phi(nb):
    p, std = 0.3, 0.2
    result = torch.empty(nb).normal_(0, std)
    result = result + torch.sign(torch.rand(result.size()) - p) / 2
    return result

######################################################################

# START_LOG_PROBA
def LogProba(x, ldj):
    log_p = ldj - 0.5 * (x**2 + math.log(2*pi))
    return log_p
# END_LOG_PROBA

######################################################################

# START_MODEL
class PiecewiseLinear(nn.Module):
    def __init__(self, n_conditions, xmin=0., xmax=20, nb=1000, xaverage_update_rate=0.25, fix_xmin=None):
        super().__init__()
        self.xmin = xmin
        if fix_xmin is not None:
            self.xmin = fix_xmin
        self.xmax = xmax
        self.xmin_avg = None
        self.xmax_avg = None
        self.fix_xmin = fix_xmin
        self.xaverage_update_rate = xaverage_update_rate
        self.nb = nb
        
        self.grad_broken = False #for debugging
        self.debug = False

        #self.input_batchnorm = nn.BatchNorm1d(n_conditions)
        #self.xi = nn.Parameter(torch.empty(nb + 1).normal_(mu, 1e-4))
        self.condition_net = torch.nn.Sequential(
            torch.nn.Linear(n_conditions, 128),
            torch.nn.ELU(), #keep it in check here, more stable
            torch.nn.Linear(128, 64),
            torch.nn.ELU(), #keep it in check here, more stable
            torch.nn.Linear(64, 64),
            torch.nn.ELU(), #keep it in check here, more stable
            torch.nn.Linear(64, nb + 1 + 1 + 1) #nb + 1 is original, 1  more for alpha, 1 more for scale
        )

    #debugging 
    def create_hook(self, layer_name):
        def hook(grad):
            if torch.isnan(grad).any() or torch.isinf(grad).any():
                #write the grad tensor to a file, focus on those entries in the first batch dimension
                #that contain nans or infs; use np.savetxt to write the tensor to a file
                mask = torch.isnan(grad) | torch.isinf(grad)
                #make the mask a B x 1 tensor in all times if any of the nested dimensions are nan or inf for that batch element
                mask = mask.any(dim=1, keepdim=False)
                np.savetxt(f'error_surrogate_{layer_name}.txt', grad[mask].detach().cpu().numpy())
                #save the indices selected by the mask above
                np.savetxt(f'error_surrogate_mask_{layer_name}.txt', np.arange(mask.shape[1])[mask].detach().cpu().numpy())
                self.grad_broken = True

            return grad  # Important to return the grad to not disrupt gradient flow
        return hook

    def adjust_xmin_xmax(self, xarray):
        '''
        this function adjusts xmin and xmax using a slow moving average from the xarray inputs
        each time it receives a new array. The actual xmin and xmax values are then set such that they
        cover 200% of the range of the input values (but still as moving average)

        Watch out, there are now hard-coded things that assume you have a loss function that has a minimum at zero!
        '''
        # this is the hard coded part now
        #mirror x array so that sigmas make sense
        #xarray = np.concatenate([xarray, -xarray],0)
        ##get std deviation
        #std = np.std(xarray)
        ## simply set moving average xmax at 4 sigma
        #xmax = 4. * std
        #if self.xmin_avg is None:
        #    self.xmin_avg = self.fix_xmin
        #    self.xmax_avg = xmax
        #
        #self.xmax = (1. - self.xaverage_update_rate) * self.xmax_avg + self.xaverage_update_rate * xmax
        #self.xmax_avg = self.xmax
        #
        #return  #below is the less hard coded version
        



        xmin = xarray.min().item()
        xmax = xarray.max().item()
        if self.xmin_avg is None:
            self.xmin_avg = xmin
            self.xmax_avg = xmax
        else:
            self.xmin_avg = (1. - self.xaverage_update_rate) * self.xmin_avg + self.xaverage_update_rate * xmin
            if  self.fix_xmin is not None:
                self.xmin_avg = self.fix_xmin
            self.xmax_avg = (1. - self.xaverage_update_rate) * self.xmax_avg + self.xaverage_update_rate * xmax
        
        #overshoot by 25% of the range
        self.xmin = self.xmin_avg #- 0.25 * (self.xmax_avg - self.xmin_avg)
        self.xmax = self.xmax_avg #+ 0.25 * (self.xmax_avg - self.xmin_avg)
        if  self.fix_xmin is not None:
                self.xmin = self.fix_xmin
        print("1D flow: xmin, xmax set to",self.xmin, self.xmax,'which is in exp space',np.exp(self.xmin), np.exp(self.xmax))

    def freeze_batchnorm(self):
        return
        for param in self.input_batchnorm.parameters():
            param.requires_grad = False

    def unfreeze_batchnorm(self):
        return
        for param in self.input_batchnorm.parameters():
            param.requires_grad = True

    def get_linears(self, conditions, return_slope = False):

        #conditions = self.input_batchnorm(conditions)
        xi_alpha = self.condition_net(conditions)

        xi = xi_alpha[:, :self.nb + 1] # B x nb+1
        alpha = xi_alpha[:, self.nb + 1:self.nb + 2] # B x 1
        scale = xi_alpha[:, self.nb + 2:self.nb + 3] # B x 1
        
        #for a non-conditioned model the three above would simply be learnable parameters

        #make scale positive
        xi = xi.softmax(1) #use softmax to make sure it's positive and all are connected - this seems more stable than exp
        xi = xi + 1e-4 #add a small value to avoid nans
        
        y = alpha + scale * xi.cumsum(1) # B x nb+1
        if not return_slope:
            return y
        
        #calculate the slope between the bin boundaries in each bin
        slopes = (y[:, 1:] - y[:, :-1]) / ((self.xmax - self.xmin)/self.nb)# B x nb
        return y, slopes
        
    def get_x_boundaries(self, x):
        return torch.linspace(self.xmin, self.xmax, self.nb + 1).to(x.device)

    def forward(self, x, conditions):
        return self.forward_jk(x, conditions)
    
    def invert(self, y, conditions):
        return self.invert_jk(y, conditions)
    
    def forward_jk(self, x, conditions):
        x = x.clamp(self.xmin, self.xmax) # B x 1
        #print('x',x.shape)
        y,slope = self.get_linears(conditions, return_slope=True) # B x nb+1
        #print('y',y.shape)
        #get the right bin
        i = self.nb * (x - self.xmin) / (self.xmax - self.xmin) # B
        #clamp to the right range
        i = i.clamp(0, self.nb - 1)
        #make it an index
        i = i.long() # B x 1
        #print('i',i.shape)
        # get boundaries
        y0 = y.gather(1, i) # B x 1
        #print('y0',y0.shape)
        # get the slope
        slope = slope.gather(1, i) # B x 1
        #print('slope',slope.shape)
        #x offset in bin
        x_off = self.get_x_boundaries(x)[i]
        #print('x_off',x_off.shape)
        #exit()
        # get the value
        return y0 + slope * (x - x_off) # B x 1
    
    def invert_jk(self, y, conditions, save_and_exit=False):
        
        ys,slope = self.get_linears(conditions, return_slope=True)
        #hook gradient for debugging
        if self.debug:
            ys.register_hook(self.create_hook('ys'))
            slope.register_hook(self.create_hook('slope'))
        
        #clamp y to be within the range
        y = y.clamp(ys[:, 0:1], ys[:, -1:])
        #search bin
        i = torch.searchsorted(ys, y)  - 1 # B x 1
        #clamp bin (should only be necessary for the last bin)
        i = i.clamp(0, self.nb - 1)
        # get boundaries
        y0 = ys.gather(1, i) # B x 1
        #print('y0',y0.shape)
        # get the slope
        slope = slope.gather(1, i) # B x 1
        # get x value in bin
        x_inbin = (y - y0) / slope # B x 1
        if self.debug:
            x_inbin.register_hook(self.create_hook('x_inbin'))
        # get the boundaries
        x_atboundary = self.get_x_boundaries(y)[i] # nb+1
        # add x bin boundary
        x = x_inbin + x_atboundary # B x 1
        if self.debug:
            x.register_hook(self.create_hook('x'))
        #assert for nans
        if torch.isnan(x).any() or torch.isinf(x).any() or save_and_exit:
            #print a bunch of stuff where the issues were and raise exception
            mask = torch.isnan(x) | torch.isinf(x)
            if save_and_exit:
                mask = torch.ones_like(x[:,0], dtype=torch.bool)
            with open('error_surrogate.txt', 'w') as f:
                f.write('nan/inf in surrogate x\n')
                f.write('y'+str(y[mask])+'\n\n')
                f.write('ys'+str(ys[mask])+'\n\n')
                f.write('slope'+str(slope[mask])+'\n\n')
                f.write('i'+str(i[mask])+'\n\n')
                f.write('x_inbin'+str(x_inbin[mask])+'\n\n')
                f.write('x_atboundary'+str(x_atboundary[mask])+'\n\n')
                f.write('conditions'+str(conditions[mask])+'\n\n')
                f.write('x'+str(x[mask])+'\n\n')
                f.write('x_min'+str(self.xmin)+'\n\n')
                f.write('x_max'+str(self.xmax)+'\n\n')
                f.write('nb'+str(self.nb)+'\n\n')
            raise Exception('nan/inf in surrogate x')
        return x



    def forward_ff(self, x, conditions):
        '''
        original implementation:
        x torch.Size([100]) : B
        y torch.Size([1002]) : nb+1
        u torch.Size([100]) : B
        n torch.Size([100]) : B
        a torch.Size([100]) : B
        out torch.Size([100]): B
        '''
        #clamp x to be within xmin and xmax
        x = x.clamp(self.xmin, self.xmax)
        #print("xi.shape, x.shape",xi.shape, x.shape)  # B x nb+1
        y = self.get_linears(conditions) # 0 -> 1 # B x nb+1
        #print("y.shape",y.shape)
        u = self.nb * (x - self.xmin) / (self.xmax - self.xmin) # B
        #print("u.shape",u.shape) 
        n = u.long().clamp(0, self.nb - 1) # B
        #print("n.shape",n.shape)
        a = (u - n).clamp(0, 1) # B
        #print("a.shape",a.shape)
        y0 = y.gather(1, n)  # Gather y values in dim 1 at indices n
        y1 = y.gather(1, n + 1)  # Gather y values in dim 1 at indices n + 1

        # now we need to use the right batch elment in y
        out = (1 - a) * y0 + a * y1
        
        return out
# END_MODEL

    def invert_ff(self, y, conditions): #FIXME also w.r.t. dimensions
        
        # Calculate ys using the cumulative sum of the exponential of xi
        ys = self.get_linears(conditions)
        
        yy = y.view(-1, 1)
        k = torch.arange(self.nb, device=y.device).view(1, -1)
        
        # Ensure y values are within the valid range
        # assert (y >= ys[:, 0]).all() and (y <= ys[:, -1]).all()
        # change w.r.t. Francois code: instead of the assert, just clip the values accordingly (ok here)
        yy = torch.clamp(yy, ys[:, 0:1].min(), ys[:, -2:-1].max())

        yk = ys[:, :-1]
        ykp1 = ys[:, 1:]
        
        # Create masks to identify the correct intervals
        masks = (yy >= yk) & (yy < ykp1)
        
        y_frac = (yy - yk) / (ykp1 - yk)
        #set to zero if nans
        y_frac[torch.logical_or(torch.isnan(y_frac), torch.isinf(y_frac))] = 0.
        # Calculate the inverse transformation within the identified intervals
        x_scale = (self.xmax - self.xmin) / self.nb
        x = self.xmin + x_scale * ((masks.float() * (k + y_frac)).sum(dim=1, keepdim=True))
        if torch.isnan(x).any() or torch.isinf(x).any():
            print('nan/inf in surrogate x')
            #write the tensors to a file error_surrogate.txt
            with open('error_surrogate.txt', 'w') as f:
                f.write('x: '+str(x) +'\n\n')
                f.write('y: '+str(y) +'\n\n')
                f.write('yy: '+str(yy) +'\n\n')
                f.write('yk: '+str(yk) +'\n\n')
                f.write('ykp1: '+str(ykp1) +'\n\n')
                f.write('y_frac: '+str(y_frac) +'\n\n')
                f.write('masks: '+str(masks) +'\n\n')
                f.write('k: '+str(k) +'\n\n')
                f.write('conditions: '+str(conditions) +'\n\n')
                f.write('xi: '+str(xi) +'\n\n')
                f.write('ys: '+str(ys) +'\n\n')
                f.write('x_scale: '+str(x_scale) +'\n\n')
                f.write('x_min: '+str(self.xmin) +'\n\n')
                f.write('x_max: '+str(self.xmax) +'\n\n')
                f.write('nb: '+str(self.nb) +'\n\n')
                f.write('alpha: '+str(self.alpha) +'\n\n')
                exit()
            
        return x
    


#nb_samples = 25000
#nb_epochs = 250
#batch_size = 100
#
#model = PiecewiseLinear(nb = 1001, xmin = -4, xmax = 4)
## model = SumOfSigmoids(nb = 51, xmin = -4, xmax = 4)
#
## print(model(torch.linspace(-10, 10, 25)))
#
## exit(0)
#
## print('** TESTING WITH POSITIVE POLYNOMIAL!!!!')
## model = PositivePolynomial(degree = 16)
#
#train_input = sample_phi(nb_samples)
#
#optimizer = torch.optim.Adam(model.parameters(), lr = 1e-4)
#
#for k in range(nb_epochs):
#    acc_loss = 0
#
## START_OPTIMIZATION
#    for input in train_input.split(batch_size):
#        input.requires_grad_()
#        output = model(input)
#
#        derivatives, = autograd.grad(
#            output.sum(), input,
#            retain_graph = True, create_graph = True
#        )
#
#        loss = ( 0.5 * (output**2 + math.log(2*pi)) - derivatives.log() ).mean()
#
#        optimizer.zero_grad()
#        loss.backward()
#        optimizer.step()
## END_OPTIMIZATION
#
#        acc_loss += loss.item()
#    if k%10 == 0: print(k, loss.item())
#
######################################################################
#
#input = torch.linspace(-3, 3, 175)
#
#mu = phi(input)
#mu_N = torch.exp(LogProba(input, 0))
#
#input.requires_grad_()
#output = model(input)
#
#grad = autograd.grad(output.sum(), input)[0]
#mu_hat = LogProba(output, grad.log()).detach().exp()
#
######################################################################
# FIGURES
#
#input = input.detach().numpy()
#output = output.detach().numpy()
#mu = mu.numpy()
#mu_hat = mu_hat.numpy()
#
######################################################################
#
#fig = plt.figure()
#ax = fig.add_subplot(1, 1, 1)
## ax.set_xlim(-5, 5)
## ax.set_ylim(-5, 5)
## ax.set_aspect('equal')
## ax.axis('off')
#
#ax.plot(input, output, '-', color = 'tab:red')
#
#filename = 'miniflow_mapping.pdf'
#print(f'Saving {filename}')
#fig.savefig(filename, bbox_inches = 'tight')
#
# plt.show()

######################################################################
#
#green_dist = '#bfdfbf'
#
#fig = plt.figure()
#ax = fig.add_subplot(1, 1, 1)
## ax.set_xlim(-4.5, 4.5)
## ax.set_ylim(-0.1, 1.1)
#lines = list(([(x_in.item(), 0), (x_out.item(), 0.5)]) for (x_in, x_out) in zip(input, output))
#lc = mc.LineCollection(lines, color = 'tab:red', linewidth = 0.1)
#ax.add_collection(lc)
#ax.axis('off')
#
#ax.fill_between(input,  0.52, mu_N * 0.2 + 0.52, color = green_dist)
#ax.fill_between(input, -0.30, mu   * 0.2 - 0.30, color = green_dist)
#
#filename = 'miniflow_flow.pdf'
#print(f'Saving {filename}')
#fig.savefig(filename, bbox_inches = 'tight')
#
# plt.show()

######################################################################
#
#fig = plt.figure()
#ax = fig.add_subplot(1, 1, 1)
#ax.axis('off')
#
#ax.fill_between(input, 0, mu, color = green_dist)
## ax.plot(input, mu, '-', color = 'tab:blue')
## ax.step(input, mu_hat, '-', where = 'mid', color = 'tab:red')
#ax.plot(input, mu_hat, '-', color = 'tab:red')
#
#filename = 'miniflow_dist.pdf'
#print(f'Saving {filename}')
#fig.savefig(filename, bbox_inches = 'tight')
#
# plt.show()

######################################################################
#
#fig = plt.figure()
#ax = fig.add_subplot(1, 1, 1)
#ax.axis('off')
#
## ax.plot(input, mu, '-', color = 'tab:blue')
#ax.fill_between(input, 0, mu, color = green_dist)
## ax.step(input, mu_hat, '-', where = 'mid', color = 'tab:red')
#
#filename = 'miniflow_target_dist.pdf'
#print(f'Saving {filename}')
#fig.savefig(filename, bbox_inches = 'tight')
#
# plt.show()

######################################################################
#
#if hasattr(model, 'invert'):
#    z = torch.randn(200)
#    z = z[(z > -3) * (z < 3)]
#
#    x = model.invert(z)
#
#    fig = plt.figure()
#    ax = fig.add_subplot(1, 1, 1)
#    ax.set_xlim(-4.5, 4.5)
#    ax.set_ylim(-0.1, 1.1)
#    lines = list(([(x_in.item(), 0), (x_out.item(), 0.5)]) for (x_in, x_out) in zip(x, z))
#    lc = mc.LineCollection(lines, color = 'tab:red', linewidth = 0.1)
#    ax.add_collection(lc)
#    # ax.axis('off')
#
#    # ax.fill_between(input,  0.52, mu_N * 0.2 + 0.52, color = green_dist)
#    # ax.fill_between(input, -0.30, mu   * 0.2 - 0.30, color = green_dist)
#
#    filename = 'miniflow_synth.pdf'
#    print(f'Saving {filename}')
#    fig.savefig(filename, bbox_inches = 'tight')
#
#    # plt.show()



###################################################################### interface code to make it a drop-in replacement for the original code


#DEBUG
torch.autograd.set_detect_anomaly(True)

class Surrogate(torch.nn.Module):
    def __init__(self, n_detector_parameters, n_true_plus_context_inputs, n_out_parameters, n_time_steps, betas=(1e-4, 0.02),
                 ):
        super(Surrogate, self).__init__()
        assert n_out_parameters == 1 #this is the whole point of a one-valued output

        self.n_detector_parameters = n_detector_parameters
        self.model = PiecewiseLinear(n_conditions = n_detector_parameters + n_true_plus_context_inputs, #true context is energy etc,
                                     xmin = 0., xmax = 30., #here the reco stuff is normalised so that should (TM) work
                                     nb=64,
                                     fix_xmin = None
                                     )
        

        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=0.0001)
        self.n_reco_parameters = n_out_parameters # called n_reco_parameters for consistency with the original code
        self.device = torch.device('cuda')
        self.n_time_steps = None #compatibility
        self.is_flow_model = True #just for checks



    

    def save(self, path):
        torch.save(self.model.state_dict(), path)

    def load(self, path):
        self.model.load_state_dict(torch.load(path))

    def adjust_xmin_xmax(self, xarray):#just forward for convenience
        # we are learning the log of the loss, so we need to take the log here
        self.model.adjust_xmin_xmax(np.log(np.where(xarray <= 0, 1e-20, xarray))) 

    @property
    def x_min(self):
        return np.exp(self.model.xmin)
    @property  
    def x_max(self):
        return np.exp(self.model.xmax)#the surrogate model lives in log space

    def forward(self, noise, detector_parameters, true_inputs, true_context, reco_step_inputs=None, time_step=None):
        assert time_step is None #make sure it's not used wrongly
        assert reco_step_inputs is None #make sure it's not used wrongly
        
        all_cond = torch.cat([detector_parameters, true_inputs, true_context], dim=1)
        #create noise with right batch size
        
        return self.model(noise, all_cond)
    
    def to(self, device = None):
        if device is None:
            device = self.device
        super(Surrogate, self).to(device)
        self.model.to(device)
        return self
    
    def create_noisy_input(self, nb):
        return torch.randn(nb, 1).to(self.device)
    
    
    def sample(self, detector_parameters, true_inputs, true_context):
        noise = self.create_noisy_input(true_inputs.shape[0]).to(self.device)
        # clip at 4 sigma
        noise = torch.clamp(noise, -3., 3.)#stability
        #
        if len(detector_parameters.shape) == 1:
            detector_parameters = detector_parameters.repeat(true_inputs.shape[0],1)
        conditions = torch.cat([detector_parameters, true_inputs, true_context], dim=1)
        log_loss = self.model.invert(noise, conditions)
        return torch.exp(log_loss)

    def sample_forward(self, detector_parameters, true_inputs, true_context):
        return self.sample(detector_parameters, true_inputs, true_context)
    
    def train_model(self, surrogate_dataset, batch_size, n_epochs, lr):
        # train the surrogate model
        train_loader = DataLoader(surrogate_dataset,  batch_size=batch_size, shuffle=True)
        # set the optimizer
        for param_group in self.optimizer.param_groups:
            param_group['lr'] = lr
        self.to(self.device)
        
        self.train()
        print('training surrogate on ',len(train_loader), 'batches')

        for epoch in range(n_epochs):
            fail_count = 0
            for batch_idx, (detector_parameters, true_inputs, true_context, reco_result) in enumerate(train_loader):
                # this needs to be adapted since it is a diffusion model. so the noise loop needs to be in here
                # the noise loop is the same as in the generator

                if fail_count > 0.5 * len(train_loader):
                    print('too many failed steps (>50%), exiting')
                    raise Exception('too many failed steps (>50%) in the surrogate training, exiting.')
                
                detector_parameters = detector_parameters.to(self.device)
                true_inputs = true_inputs.to(self.device)
                reco_step_inputs = reco_result.to(self.device)
                
                #add an epsilon to avoid log(0)
                reco_step_inputs = torch.where(reco_step_inputs == 0, 1e-16, reco_step_inputs)
                ## we learn the log of the loss (reco_step_inputs)
                reco_step_inputs = torch.log(reco_step_inputs)


                true_context = true_context.to(self.device)
                reco_step_inputs.requires_grad_()

                model_out = self(reco_step_inputs,detector_parameters, true_inputs, true_context)

                derivatives, = autograd.grad(
                            model_out.sum(), reco_step_inputs,
                            retain_graph = True, create_graph = True
                            )
                #this loss can be a bit instable
                loss = 0.5 * (model_out**2 + math.log(2*pi)) - torch.log(derivatives + 1e-16)
                #stable_loss = 0.5 * (model_out**2 + math.log(2*pi)) - torch.log(torch.abs(derivatives) + 1e-4)
                #stable_loss +=  torch.relu(-derivatives)**2 #push them up with relu - derivatives
                #protection: replace with less accurate version where it matters
                #loss = torch.where( torch.logical_or(torch.isnan(loss), torch.isinf(loss)) , stable_loss, loss)
                #remove some remaining nans, this can happen in particular in the beginning
                loss = loss[torch.logical_not(torch.logical_or(torch.isnan(loss), torch.isinf(loss)))]
                
                loss = loss.mean()
                #print(loss)

                self.optimizer.zero_grad()
                loss.backward()
                if self.model.grad_broken:
                    print('gradient broken, exiting')
                    raise Exception('gradient broken in the surrogate training, exiting.')
                #add some stability by clipping the gradients
                torch.nn.utils.clip_grad_value_(self.model.parameters(), 1.)
                
                self.optimizer.step()
                derivatives = derivatives.detach()
                loss = loss.detach()
                reco_step_inputs = reco_step_inputs.detach()


            print('Surrogate Epoch: {} \tLoss (negative is ok): {:.8f}'.format(
                        epoch, loss.item()))
        #detach all parameters and the model itself so the graph is destroyed

        self.eval()
        return loss.item()

    def apply_model_in_batches(self, dataset, batch_size, oversample=1):
        '''
        one to one copy of the original function, no changes needed
        '''

        self.to()
        self.eval()
        # create a dataloader for the dataset
        data_loader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
        # create a tensor to store the results
        results = []
        reco = []
        true = []
        for i_o in range(oversample):
            # loop over the batches
            for batch_idx, (detector_parameters, true_inputs, true_context, reco_inputs) in enumerate(data_loader): # the reco is not needed as it is generated here
                
                print(f'batch {batch_idx} of {len(data_loader)}', end='\r')
                detector_parameters = detector_parameters.to(self.device)
                true_inputs = true_inputs.to(self.device)
                true_context = true_context.to(self.device)
                reco_inputs = reco_inputs.to(self.device)
                # apply the model
                reco_surrogate = self.sample(detector_parameters, true_inputs, true_context)

                #we have learned the log of the loss, so we need to unlog it

                #un_normalise all to physical values. but the surrogate output is now already the loss FLOW_ADJUSTMENT
                #reco_surrogate = dataset.unnormalise_target(reco_surrogate)
                reco_inputs = dataset.unnormalise_target(reco_inputs)
                true_inputs = dataset.unnormalise_target(true_inputs)
                # store the results
                results.append(reco_surrogate.detach().to('cpu'))
                reco.append(reco_inputs.detach().to('cpu'))
                true.append(true_inputs.detach().to('cpu'))

        return torch.concat(results, 0), torch.concat(reco, 0), torch.concat(true, 0)
