import multiprocessing
import time


import matplotlib.pyplot as plt
import numpy as np
import pickle


import threading
import queue
import time

def plot_evolution_with_panels(evolution_file, outpath, show=False):
    '''
    Read the evolution file and plot the evolution of parameters and losses in two panels.
    '''
    # Load the evolution data
    with open(evolution_file, 'rb') as f:
        evolution = pickle.load(f)
    
    if len(evolution) == 0:
        print("No evolution data found.")
        return
    
    # Transpose the parameters for plotting
    first_step = evolution[0]  # Use the first step to get the structure
    ps_dict_keys = first_step['parameters'].keys()
    pltdict = {k: [] for k in ps_dict_keys}
    
    # Populate the plotting dictionary with parameter values over time
    for step in evolution:
        for k in pltdict.keys():
            pltdict[k].append(step['parameters'][k])
    
    #print(pltdict.keys(), pltdict['thickness_absorber_78'])
    #exit()
    # Set up the figure with two subplots (top and bottom)
    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(10, 10), gridspec_kw={'height_ratios': [2, 1]})
    
    # Plot the parameters evolution on the top panel (ax1)
    for k in pltdict.keys():
        style = '--' if 'absorber' in k else '-'
        ax1.plot(range(len(pltdict[k])), np.array(pltdict[k]), 
                 label=k, linestyle=style)
    
    # Set up legend and labels for the top plot
    ax1.set_ylabel('Parameter Values')
    ax1.set_xlabel('Step')
    
    # Plot the losses (including reco_val_loss) on the bottom panel (ax2)
    ax2.plot(range(len(evolution)), [step['losses'] for step in evolution], label='loss', 
             color='black', linewidth=2, alpha=0.2)
    ax2.plot(range(len(evolution)), [step['reco_losses'] for step in evolution], label='reco loss', 
             color='black', linewidth=2, alpha=0.2, linestyle='--')
    
    reco_val_loss = [step['reco_val_loss'] for step in evolution]
    #replace None with NaN and make it an array
    reco_val_loss = np.array([np.nan if x is None else x for x in reco_val_loss])
    print(reco_val_loss)
    ax2.plot(range(len(evolution)), reco_val_loss, label='reco val loss', 
             color='black', marker='o')
    
    # Log scale for losses
    ax2.set_yscale('log')
    ax2.set_ylabel('Loss Values')
    ax2.set_xlabel('Step')
    
    # Make space for a larger legend on the right
    plt.subplots_adjust(right=0.7)
    
    # Move the legends outside of the plots to the right
    ax1.legend(loc='center left', bbox_to_anchor=(1.05, 0.15), prop={'size': 4}, ncol=2)
    ax2.legend()#put it in the plot
    
    # Save the resulting figure
    plt.savefig(outpath + '_evolution_split_panels.png')
    plt.savefig(outpath + '_evolution_split_panels.pdf')
    if show:
        plt.show()
    plt.close()

class ThreadingPool:
    def __init__(self, max_workers):
        self.max_workers = max_workers  # Limit on the number of concurrent threads
        self.task_queue = queue.Queue()  # Queue to hold tasks
        self.results = []  # To store the results of the tasks
        self.threads = []  # List of currently running threads
        self.lock = threading.Lock()  # Lock to manage shared resources

    def add_task(self, task, *args, **kwargs):
        """Adds a task to the queue."""
        self.task_queue.put((task, args, kwargs))

    def _worker(self):
        """The worker function that will process tasks from the queue."""
        while True:
            try:
                task, args, kwargs = self.task_queue.get_nowait()
            except queue.Empty:
                break  # Exit when the queue is empty

            try:
                result = task(*args, **kwargs)  # Run the task with its arguments
                with self.lock:
                    self.results.append(result)  # Store the result in a thread-safe way
            finally:
                self.task_queue.task_done()  # Mark the task as done

    def run(self):
        """Launches the threads and processes tasks in the queue."""
        # Keep launching threads until the queue is empty
        while not self.task_queue.empty():
            if len(self.threads) < self.max_workers:
                # Start a new thread if we haven't reached the max_workers limit
                thread = threading.Thread(target=self._worker)
                thread.start()
                self.threads.append(thread)

            # Remove completed threads from the list
            self.threads = [t for t in self.threads if t.is_alive()]

            # Avoid busy waiting by sleeping for a short time
            time.sleep(0.1)

        # Wait for all threads to finish
        for thread in self.threads:
            thread.join()

    def get_results(self):
        """Returns the results of the completed tasks."""
        return self.results




# Example usage of the ThreadingPool class
#def main():
#    pool = ThreadingPool(max_workers=3)
#
#    # Add tasks to the pool
#    for i in range(10):
#        pool.add_task(example_task, i, duration=2)
#
#    # Run the pool (this will start the threads and process the tasks)
#    pool.run()
#
#    # Get the results (order does not matter)
#    results = pool.get_results()
#    print("All tasks completed.")
#    for result in results:
#        print(result)
#
#if __name__ == "__main__":
#    main()
